/* Sol's 2d gl basecode
 *  version 1.0, December 16th, 2007
 *
 *  Copyright (C) 2007 Jari KOmppa
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty.  In no event will the authors be held liable for any damages
 *  arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any purpose,
 *  including commercial applications, and to alter it and redistribute it
 *  freely, subject to the following restrictions:
 *
 *  1. The origin of this software must not be misrepresented; you must not
 *     claim that you wrote the original software. If you use this software
 *     in a product, an acknowledgment in the product documentation would be
 *     appreciated but is not required.
 *  2. Altered source versions must be plainly marked as such, and must not be
 *     misrepresented as being the original software.
 *  3. This notice may not be removed or altered from any source distribution.
 *
 * Jari Komppa http://iki.fi/sol/
 *
 *************************************
 * 
 */

#include "toolkit.h"

#define TITLE "Program Title 1.0"

// How big a tick difference is considered 'time warp', i.e. skip the time
// (to avoid physics blowing up)
#define TICK_TIMEWARP 1000
// Physics ticks per second
#define PHYSICS_FPS 100
// For windowed mode
#define DESIRED_WINDOW_WIDTH 800
// For windowed mode
#define DESIRED_WINDOW_HEIGHT 600
// Desired aspect ratio. 640x480 = 4/3. Comment out to use whatever the desktop is set.
#define DESIRED_ASPECT ((float)(DESIRED_WINDOW_WIDTH)/(float)(DESIRED_WINDOW_HEIGHT))

//#define FULLSCREEN_BY_DEFAULT

#define USE_PERFCOUNTERS

int tex_logo, tex_font;
int gKeyState = 0;
int gLastTick = 0;
int gScreenWidth = 0;
int gScreenHeight = 0;

ACFont fn, fn14;

void handle_key(int keysym, int down)
{
    switch(keysym)
    {
    case SDLK_ESCAPE:
        if (!down)
        {
            exit(0);
        }
        break;
    case SDLK_UP:
    case SDLK_w:
        if (down) gKeyState |= KEY_UP; else gKeyState &= ~KEY_UP;
        break;
    case SDLK_DOWN:
    case SDLK_s:
        if (down) gKeyState |= KEY_DOWN; else gKeyState &= ~KEY_DOWN;
        break;
    case SDLK_LEFT:
    case SDLK_a:
        if (down) gKeyState |= KEY_LEFT; else gKeyState &= ~KEY_LEFT;
        break;
    case SDLK_RIGHT:
    case SDLK_d:
        if (down) gKeyState |= KEY_RIGHT; else gKeyState &= ~KEY_RIGHT;
        break;
    case SDLK_RCTRL:
    case SDLK_LCTRL:
        if (down) gKeyState |= KEY_FIRE;
        break;
    }
}

void process_events()
{
    SDL_Event event;

    while (SDL_PollEvent(&event)) 
    {
        switch (event.type) 
        {
        case SDL_KEYDOWN:
            handle_key(event.key.keysym.sym, 1);
			// If a key is pressed, report it to the widgets
			gUIState.keyentered = event.key.keysym.sym;
			gUIState.keymod = event.key.keysym.mod;
			// if key is ASCII, accept it as character input
			if ((event.key.keysym.unicode & 0xFF80) == 0)
				gUIState.keychar = event.key.keysym.unicode & 0x7f;				
            break;
        case SDL_KEYUP:
            handle_key(event.key.keysym.sym, 0);
            break;
        case SDL_MOUSEMOTION:
			// update mouse position
			gUIState.mousex = (event.motion.x * (float)DESIRED_WINDOW_WIDTH) / (float)gScreenWidth;
			gUIState.mousey = (event.motion.y * (float)DESIRED_WINDOW_HEIGHT) / (float)gScreenHeight;
            break;
		case SDL_MOUSEBUTTONDOWN:
			// update button down state if left-clicking
			if (event.button.button == 1)
				gUIState.mousedown = 1;
			break;
		case SDL_MOUSEBUTTONUP:
			// update button down state if left-clicking
			if (event.button.button == 1)
				gUIState.mousedown = 0;
			break;
        case SDL_QUIT:
            SDL_Quit();
            exit(0);
            break;
        }
    }
}



void physics_tick(int tick)
{
    int i;
    for (i = 0; i < MAX_PARTICLES; i++)
    {
        if (particle[i].live)
        {
            particle[i].live--;
            particle[i].x += particle[i].dirx;
            particle[i].y += particle[i].diry;
        }
    }        

    physics_popups();

    ////////////////////////////////////
    // testcode
    ////////////////////////////////////
    spawn_particle(rand() % 800, rand() % 600, tex_logo);
    if ((tick % 100) == 0) spawn_popup("Laaa..",rand()%800,rand()%600);
    ////////////////////////////////////
    // /testcode
    ////////////////////////////////////

}


static void draw_screen_ingame()
{
    int tick = SDL_GetTicks();

    int i, j;    

    if (tick - gLastTick > TICK_TIMEWARP) 
        gLastTick = tick;

    if (gLastTick >= tick)
    {
        SDL_Delay(1);
        return;
    }

    while (gLastTick < tick)
    {
        physics_tick(gLastTick);

        gLastTick += 1000 / PHYSICS_FPS;
    }

    ////////////////////////////////////
    // Rendering
    ////////////////////////////////////

    glClearColor(0.1,0.1,0.2,1.0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    ////////////////////////////////////
    // testcode
    ////////////////////////////////////

    drawsprite(tex_logo,64,64,128,128);

    for (i = 0; i < MAX_PARTICLES; i++)
    {
        if (particle[i].live)
            drawsprite(particle[i].tex, particle[i].x, particle[i].y, particle[i].live * (7 / 100.0f), particle[i].live * (7 / 100.0f), 0.5f);
    }

    for (j = 0; j < MAX_POPUPS; j++)
    {
        if (popup[j].live)
        {
            quickfont_drawstring(tex_font, popup[j].text, popup[j].x + 2, popup[j].y + 2, 0, popup[j].live / 200.0f);
            quickfont_drawstring(tex_font, popup[j].text, popup[j].x, popup[j].y, 0xffff7f, popup[j].live / 200.0f);
        }
    }

    char temp[200];
    sprintf(temp, "%d %d %d %d", gScreenWidth, gScreenHeight, gUIState.mousex, gUIState.mousey);
    quickfont_drawstring(tex_font, temp, 50,400,0xffffff,1,1);

    imgui_prepare();

    static char textfield_text[200];

    if (imgui_button(GEN_ID,fn14,"Hello1",50,150,64,40,0xff3f3f7f,0xff7f7f9f,0xff9f9fff,0xff000000))
    {
        char * hello = "Hello world!\0";
        memcpy(textfield_text,hello,strlen(hello)+1);
    }
    if (imgui_button(GEN_ID,fn14,"Hello2",50,200,64,40,0xff3f3f7f,0xff7f7f9f,0xff9f9fff,0xff000000))
    {
        char * hello = "Hello Starsystem!\0";
        memcpy(textfield_text,hello,strlen(hello)+1);
    }
    if (imgui_button(GEN_ID,fn14,"Hello3",50,250,64,40,0xff3f3f7f,0xff7f7f9f,0xff9f9fff,0xff000000))
    {
        char * hello = "Hello Galaxy!\0";
        memcpy(textfield_text,hello,strlen(hello)+1);
    }
    if (imgui_button(GEN_ID,fn14,"Hello4",50,300,64,40,0xff3f3f7f,0xff7f7f9f,0xff9f9fff,0xff000000))
    {
        char * hello = "Hello Universe!\0";
        memcpy(textfield_text,hello,strlen(hello)+1);
    }

    static int slidervalue = 25;
    imgui_slider(GEN_ID,760,50,24,400,0xff3f3f7f,0xff7f7f9f,0xff9f9fff,100,slidervalue);
    
    imgui_textfield(GEN_ID,fn,100,450,600,44,textfield_text,100,0xff3f3f7f,0xff7f7f9f,0xff9f9fff,0xff000000);

    imgui_finish();

    drawrect(gUIState.mousex,gUIState.mousey,16,16,0xff00ff00);

    float w, h;
    fn.stringmetrics(textfield_text,w,h,slidervalue * 1.1 + 10);
    fn.drawstring(textfield_text,(800 - w) / 2,(600 - h) / 2,0x7f000000,slidervalue*1.1 + 10);
    fn.stringmetrics(textfield_text,w,h,slidervalue + 10);
    fn.drawstring(textfield_text,(800 - w) / 2,(600 - h) / 2,0x7fffffff,slidervalue + 10);

    ////////////////////////////////////
    // /testcode
    ////////////////////////////////////
    

#ifdef USE_PERFCOUNTERS
    ////////////////////////////////////
    // performance counter
    ////////////////////////////////////

#define PERF_FRAMES 50
    drawrect(8,gScreenHeight - 34 - 40,210,40 + 20,0x3f0000ff);

    drawrect(8,gScreenHeight - 34 - 40,210,1,0x3f0000ff);
    drawrect(8,gScreenHeight - 34 - 40,1,40 + 20,0x3f0000ff);
    drawrect(8,gScreenHeight - 34 + 19,210,1,0x3f0000ff);
    drawrect(8 + 209,gScreenHeight - 34 - 40,1,40 + 20,0x3f0000ff);

    static int perf_idx = 0;
    static int perf_data[PERF_FRAMES];
    perf_data[perf_idx] = tick;
    char perf_temp[200];
    sprintf(perf_temp, "%3.3f mspf (%3.3f fps)", (float)(perf_data[perf_idx] - perf_data[(perf_idx+1) % PERF_FRAMES]) / PERF_FRAMES,
                                                 1000.0f / ((float)(perf_data[perf_idx] - perf_data[(perf_idx+1) % PERF_FRAMES]) / PERF_FRAMES));
    perf_idx++;
    perf_idx %= PERF_FRAMES;
    quickfont_drawstring(tex_font, perf_temp, 16,gScreenHeight - 32,0xffffff,1,1);
    int perf_i;
    for (perf_i = 0; perf_i < PERF_FRAMES - 1; perf_i++)
    {
        float h = (perf_data[(perf_idx + perf_i + 1) % PERF_FRAMES] - perf_data[(perf_idx + perf_i) % PERF_FRAMES]);
        drawrect(perf_i * 4 + 16, gScreenHeight - 34 - h, 3, h, 0x7fff0000);
    }

    ////////////////////////////////////
    // /performance counter
    ////////////////////////////////////
#endif
    SDL_Delay(10);
	SDL_GL_SwapBuffers();
}



void draw_screen()
{
    draw_screen_ingame();
}


int main (int argc, char** args)
{
    const SDL_VideoInfo *info = NULL;
    int bpp = 0;
    int flags = 0;

    gVisualRand.init_genrand(0xc0cac01a);

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) < 0) 
    {
        fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

    info = SDL_GetVideoInfo();

    if (!info) 
    {
        fprintf(stderr, "Video query failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

#ifdef _DEBUG
    int fsflag = 0;
#else
#if FULLSCREEN_BY_DEFAULT
    int fsflag = 1;
#else
    int fsflag = 0;
#endif
#endif

    if (argc > 1) fsflag = !fsflag;

    if (fsflag) 
    {
        gScreenWidth = info->current_w;
        gScreenHeight = info->current_h;
        bpp = info->vfmt->BitsPerPixel;
        flags = SDL_OPENGL | SDL_FULLSCREEN;
    }
    else
    {
        gScreenWidth = DESIRED_WINDOW_WIDTH;
        gScreenHeight = DESIRED_WINDOW_HEIGHT;
        bpp = info->vfmt->BitsPerPixel;
        flags = SDL_OPENGL;
    }

    SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

    if (SDL_SetVideoMode(gScreenWidth, gScreenHeight, bpp, flags) == 0) 
    {
        fprintf( stderr, "Video mode set failed: %s\n", SDL_GetError());
        SDL_Quit();
        exit(0);
    }

    SDL_WM_SetCaption(TITLE " - http://iki.fi/sol/", NULL);
    
#ifdef DESIRED_ASPECT
    float aspect = DESIRED_ASPECT;
    if ((float)gScreenWidth / gScreenHeight > aspect)
    {
        float realx = gScreenHeight * aspect;
        float extrax = gScreenWidth - realx;

        glViewport( extrax / 2, 0, realx, gScreenHeight );
    }
    else
    {
        float realy = gScreenWidth / aspect;
        float extray = gScreenHeight - realy;

        glViewport( 0, extray / 2, gScreenWidth, realy );
    }
#else
    glViewport( 0, 0, gScreenWidth, gScreenHeight );
#endif


	// For imgui - Enable keyboard repeat to make sliders more tolerable
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
	// For imgui - Enable keyboard UNICODE processing for the text field.
	SDL_EnableUNICODE(1);


    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
#ifdef DESIRED_ASPECT
	gluOrtho2D(0,DESIRED_WINDOW_WIDTH,DESIRED_WINDOW_HEIGHT,0);
#else
    gluOrtho2D(0,gScreenWidth,gScreenHeight,0);
#endif

    tex_logo = load_texture("spherebase.png");
    tex_font = load_texture("fona.png");

    fn.load("vera64.fnt");
    fn14.load("vera14.fnt");

    SDL_ShowCursor(1);
    
    while (1) 
    {
        process_events();
        draw_screen();
    }

    return 0;
}
